@@ -2,6 +2,8 @@ ALLOW_CERTIFICATE_PUSHING = False
 MORANGO_SERIALIZE_BEFORE_QUEUING = True
 MORANGO_DESERIALIZE_AFTER_DEQUEUING = True
 MORANGO_DISALLOW_ASYNC_OPERATIONS = False
+MORANGO_DISABLE_FSIC_V2_FORMAT = False
+MORANGO_INSTANCE_INFO = {}
 MORANGO_INITIALIZE_OPERATIONS = (
     "morango.sync.operations:InitializeOperation",
     "morango.sync.operations:LegacyNetworkInitializeOperation",
@@ -27,7 +29,7 @@ MORANGO_TRANSFERRING_OPERATIONS = (
 MORANGO_DEQUEUE_OPERATIONS = (
     "morango.sync.operations:ProducerDequeueOperation",
     "morango.sync.operations:ReceiverDequeueOperation",
-    "morango.sync.operations:LegacyDequeueOperation",
+    "morango.sync.operations:LegacyNetworkDequeueOperation",
     "morango.sync.operations:NetworkDequeueOperation",
 )
 MORANGO_DESERIALIZE_OPERATIONS = (
